import pandas as pd
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
from upsetplot import from_memberships
from upsetplot import plot
from typing import List, Tuple, Union
from glob import glob
import pathlib
from .config import logger

plt.rcParams["svg.fonttype"] = "none"
plt.rcParams["pdf.fonttype"] = 42


class BarcodeAnalyzer:
    def __init__(self, directory: str) -> None:
        """

        :param directory: Path to the directory that contains barcode CSV files
        """

        self.barcodes = {}
        for i in sorted(glob(f"./{directory}/*.csv")):
            name = pathlib.Path(i).name.replace(".csv", "")
            self.barcodes[name] = pd.read_csv(i, header=None, index_col=0)[1]

        if len(self.barcodes) == 0:
            logger.error(f"No barcodes found in the directory {directory}")
        else:
            self.barcodes = pd.DataFrame(self.barcodes).fillna(0).astype(int)
            self.barcodes.index.rename("barcodes", inplace=True)
            logger.info(f"{len(self.barcodes.columns)} barcode files found")

    def merge_groups(self, group_vec: List[str]) -> None:
        """

        :param group_vec:
        :return:
        """
        df = self.barcodes.copy().T
        df["group"] = group_vec
        df = df.groupby("group").min().T
        self.barcodes = df

    def make_subset(self, columns: List[str]) -> None:
        """

        :param columns:
        :return:
        """
        self.barcodes = self.barcodes[columns]
        self.barcodes = self.barcodes[self.barcodes.sum(axis=1) != 0]

    def total_barcodes(self) -> pd.Series:
        df = self.barcodes.copy()
        df[df > 0] = 1
        return df.sum()

    def all_overlaps(self, count: bool = True) -> Union[int, pd.DataFrame]:
        df = self.barcodes.copy()
        df[df > 0] = 1
        xdf = df.sum(axis=1) == len(df.columns)
        if count:
            return xdf.sum()
        else:
            return df[xdf]

    def calc_pairwise_overlap(self, corrected: bool = False) -> pd.DataFrame:
        """

        :param corrected:
        :return:
        """
        overlap = {}
        df = self.barcodes.copy()
        df[df > 0] = 1
        for i in df.columns:
            overlap[i] = {}
            for j in df.columns:
                if i != j:
                    overlap[i][j] = ((df[i] + df[j]) == 2).sum()
                    if corrected:
                        overlap[i][j] = overlap[i][j] / ((df[i] + df[j]) > 0).sum()
                else:
                    overlap[i][j] = ((df[i] > 0) & (df.sum(axis=1) == 1)).sum()
                    if corrected:
                        overlap[i][j] = 1
        return pd.DataFrame(overlap)

    def calc_multi_overlap(self, show_missing: bool = False) -> pd.DataFrame:
        df = self.barcodes.copy()
        df[df > 0] = 1
        if show_missing is False:
            df = df[df.sum(axis=1) > 0]
        xdf = (
            pd.Series(["".join(map(str, x)) for x in df.values])
            .value_counts()
            .sort_index()
        )
        return from_memberships(
            [
                list(df.columns[np.array(list(map(int, x))).astype(bool)])
                for x in xdf.index
            ],
            data=xdf.values,
        )

    def calc_weighted_overlap(self) -> pd.DataFrame:
        """

        :return:
        """
        xdf = self.barcodes / self.barcodes.sum()
        w_o = {}
        for i in xdf:
            w_o[i] = {}
            for j in xdf:
                a = xdf[i]
                b = xdf[j]
                w_o[i][j] = ((a - b) ** 2).fillna(0).sum()
        w_o = pd.DataFrame(w_o)
        return 1 - w_o / w_o.max().max()

    def calc_percentage_overlap(self) -> pd.DataFrame:
        """

        :return:
        """
        overlap = self.calc_pairwise_overlap()
        return 100 * overlap / overlap.sum()

    def plot_stacked(
        self,
        fig_size: Tuple[int, int] = (7, 5),
        save_name: str = None,
        rotation: float = 70,
    ) -> None:
        """

        :param fig_size:
        :param save_name:
        :param rotation:
        :return:
        """
        p_overlap = self.calc_percentage_overlap().cumsum().T
        fig, ax = plt.subplots(1, 1, figsize=fig_size)
        xind = list(range(p_overlap.shape[0]))
        for i in p_overlap.columns[::-1]:
            ax.bar(xind, p_overlap[i].values, label=i)
        ax.set_xticks(xind)
        ax.set_xticklabels(p_overlap.index, rotation=rotation)
        ax.legend(loc=(1.1, 0.5))
        ax.set_ylabel("% barcode overlap")
        plt.tight_layout()
        if save_name is not None:
            plt.savefig(save_name, dpi=300)
        plt.show()

    def plot_upset(
        self,
        fig_size: Tuple[int, int] = (7, 5),
        save_name: str = None,
        show_counts: bool = True,
        show_missing: bool = False,
    ) -> None:
        """

        :param show_missing:
        :param show_counts:
        :param fig_size:
        :param save_name:
        :return:
        """
        table = self.calc_multi_overlap(show_missing=show_missing)
        fig = plt.figure(figsize=fig_size)
        if show_counts:
            show_counts = "%d"
        else:
            show_counts = None
        plot(table, sort_by="cardinality", fig=fig, show_counts=show_counts)
        if save_name is not None:
            plt.savefig(save_name, dpi=300)
        plt.show()

    def plot_weighted_heatmap(
        self,
        fig_size: Tuple[int, int] = (7, 7),
        cmap: str = "coolwarm",
        save_name: str = None,
        robust: bool = True,
    ) -> None:
        """

        :param fig_size:
        :param cmap:
        :param save_name:
        :param robust:
        :return:
        """
        sns.clustermap(
            self.calc_weighted_overlap(), cmap=cmap, figsize=fig_size, robust=robust
        )
        if save_name is not None:
            plt.savefig(save_name, dpi=300)
        plt.show()

    def plot_overlap_heatmap(
        self,
        fig_size: Tuple[int, int] = (7, 7),
        cmap: str = "coolwarm",
        save_name: str = None,
        robust: bool = True,
    ) -> None:
        """

        :param fig_size:
        :param cmap:
        :param save_name:
        :param robust:
        :return:
        """
        xdf = self.calc_pairwise_overlap(corrected=True)
        xdf[xdf == 1] = 0
        sns.clustermap(
            self.calc_pairwise_overlap(corrected=True),
            cmap=cmap,
            figsize=fig_size,
            vmax=xdf.max().max(),
            robust=robust,
        )
        if save_name is not None:
            plt.savefig(save_name, dpi=300)
        plt.show()



import gzip
import re
import editdistance
from tqdm import tqdm
import pandas as pd
from typing import Generator, Tuple, Optional, List
import matplotlib.pyplot as plt
import numpy as np
from .config import logger

plt.rcParams["svg.fonttype"] = "none"


class BarcodeExtractor:
    def __init__(
        self,
        read1_file: str,
        read2_file: str,
        left_flank: Optional[str] = None,
        right_flank: Optional[str] = None,
        barcode_length: Optional[int] = None,
        max_dist: int = 3,
        max_read_length: int = 200,
        disable_progress_bar: bool = False,
    ):
        """
        Extract barcodes from paired-end FASTQ files. The files can be gunzip.

        :param read1_file: Read 1 FASTQ file
        :param read2_file: Read 2 FASTQ file
        :param left_flank: Left primer sequence
        :param right_flank: Right primer sequence
        :param barcode_length: Length of the barcode sequence
        :param max_dist: Maximum allowed distance between barcode sequence from match pair.
        :param max_read_length: Maximum read length
        :param disable_progress_bar: If True, then progress bars are not shown
        """
        self.r1Fn = read1_file
        self.r2Fn = read2_file
        self.leftFlank = left_flank
        self.rightFlank = right_flank
        self.barcodeLength = barcode_length
        self.maxDist = max_dist
        self.maxReadLength = max_read_length
        self.nucComp: Optional[pd.Series] = None
        self.rawCounts: Optional[pd.Series] = None
        self.disablePb = disable_progress_bar

    @staticmethod
    def _read_fastq(fn: str) -> Generator[str, None, None]:
        n = 1
        if fn.endswith(".gz"):
            to_str = lambda x: x.decode("UTF-8").rstrip("\n")
            handle = gzip.open(fn)
        else:
            to_str = lambda x: x.rstrip("\n")
            handle = open(fn)
        for line in handle:
            if n % 2 == 0 and n % 4 != 0:
                yield to_str(line)
            n += 1
        handle.close()

    def identify_flanks(
        self,
        max_frac_threshold: float = 0.6,
        flank_size: int = 6,
        n_rows: Optional[int] = None,
    ):
        """

        :param max_frac_threshold:
        :param flank_size:
        :param n_rows:
        :return:
        """
        nuc_comp = {x: {x: 0 for x in "ATGCN"} for x in range(self.maxReadLength)}
        n = 0
        for seq1 in tqdm(
            self._read_fastq(self.r1Fn),
            desc="Identifying flank sequences",
            disable=self.disablePb,
        ):
            for p, i in enumerate(seq1):
                nuc_comp[p][i] += 1
            n += 1
            if n_rows is not None and n > n_rows:
                break
        nuc_comp = pd.DataFrame(nuc_comp).T.sort_index()
        ncs = nuc_comp.sum(axis=1).values
        length_thresh = np.where(ncs > ncs[0] / 2)[0].max()
        nuc_comp = nuc_comp[:length_thresh]
        frac = nuc_comp.max(axis=1) / nuc_comp.sum(axis=1)
        p = np.where(frac < max_frac_threshold)[0]
        s, e = p[0], p[-1]
        if p.shape[0] == e - s + 1:
            self.leftFlank = "".join(nuc_comp[s - flank_size : s].idxmax(axis=1).values)
            self.rightFlank = "".join(
                nuc_comp[e + 1 : e + flank_size + 1].idxmax(axis=1).values
            )
            self.barcodeLength = len(p)
        else:
            logger.error(
                "Flank detection failed due to non-contiguous variable base composition "
            )
        self.nucComp = nuc_comp

    @staticmethod
    def _rev_comp(seq: str) -> str:
        rev_map = {"A": "T", "T": "A", "G": "C", "C": "G", "N": "N"}
        return "".join([rev_map[x] for x in seq[::-1]])

    def _extract_seq_with_flanks(
        self, seq: str, up_seq: str, down_seq: str
    ) -> Optional[str]:
        """

        :param seq:
        :param up_seq:
        :param down_seq:
        :return:
        """
        up_pos = [x.start() for x in re.finditer(up_seq, seq)]
        down_pos = [x.start() for x in re.finditer(down_seq, seq)]
        adjust_start = len(up_seq)
        barcode_len = self.barcodeLength + adjust_start
        if len(up_pos) > 0 and len(down_pos) > 0:
            valid_comb = []
            for i in up_pos:
                for j in down_pos:
                    if j - i == barcode_len:
                        valid_comb.append((i + adjust_start, j))
            if len(valid_comb) == 1:
                return seq[valid_comb[0][0] : valid_comb[0][1]]
        return None

    def _stream_reads(self) -> Generator[Tuple[str, str], None, None]:
        for seq1, seq2 in zip(self._read_fastq(self.r1Fn), self._read_fastq(self.r2Fn)):
            yield seq1, seq2
        return None

    def count_barcodes(
        self,
    ) -> None:
        """

        :return:
        """
        if (
            self.leftFlank is None
            or self.rightFlank is None
            or self.barcodeLength is None
        ):
            logger.error(
                "Flank sequence(s) and/or barcode length is unknown. Please run `identify_flanks` "
                "to detect automatically or provide them manually when calling `BarcodesPE`"
            )
            return None
        counts = {}
        n, f1, f2 = 0, 0, 0
        max_len = 0
        nuc_comp = {x: {x: 0 for x in "ATGCN"} for x in range(self.maxReadLength)}
        for seq1, seq2 in tqdm(
            self._stream_reads(), disable=self.disablePb, desc="Counting barcodes"
        ):
            if len(seq1) > max_len:
                max_len = len(seq1)
            for p, i in enumerate(seq1):
                nuc_comp[p][i] += 1
            b1, b2 = (
                self._extract_seq_with_flanks(seq1, self.leftFlank, self.rightFlank),
                self._extract_seq_with_flanks(
                    seq2,
                    self._rev_comp(self.rightFlank),
                    self._rev_comp(self.leftFlank),
                ),
            )
            if b1 is not None and b2 is not None:
                b2 = self._rev_comp(b2)
                if editdistance.eval(b1, b2) <= self.maxDist:
                    if b1 not in counts:
                        counts[b1] = 0
                    counts[b1] += 1
                else:
                    f2 += 1
            else:
                f1 += 1
            n += 1

        f1 = 100 * f1 / n
        f2 = 100 * f2 / n
        f1 = "%.2f" % f1
        f2 = "%.2f" % f2
        logger.info(
            f"{n} sequences processed. {len(counts)} unique(uncorrected) barcodes found."
        )
        logger.warning(
            f"Unable to find barcodes in {f1}% reads. {f2}% reads had too many mismatches."
        )
        if max_len > self.maxReadLength:
            logger.warning(
                f"Maximum observed read length ({max_len}) is higher than provided ({self.maxReadLength}). "
                "The nucleotide composition table is likely to be truncated."
            )
        else:
            logger.info(f"Maximum observed read length is {max_len}")
            self.maxReadLength = max_len
        self.nucComp = pd.DataFrame(nuc_comp).T.sort_index()[: self.maxReadLength]
        self.rawCounts = pd.Series(
            dict(sorted(counts.items(), key=lambda item: item[1], reverse=True))
        )

    def plot_composition(
        self,
        save_name: str = None,
        verts: List[int] = None,
        fig_size: Tuple[int, int] = (12, 3),
    ) -> None:
        """

        :param save_name:
        :param verts:
        :param fig_size:
        :return:
        """
        fig, ax = plt.subplots(1, 1, figsize=fig_size)
        self.nucComp.plot(kind="bar", stacked=True, ax=ax, cmap="Set3")
        if verts is not None:
            for i in verts:
                ax.axvline(i, color="k", lw=3.5, ls="--")
        ax.set_xlabel("Position in read")
        ax.set_ylabel("Nucleotide frequency")
        plt.tight_layout()
        if save_name is not None:
            plt.savefig(save_name, dpi=300)
            plt.close()
        else:
            plt.show()

    def plot_barcode_frequency(
        self, save_name: str = None, fig_size: Tuple[int, int] = (6, 5)
    ) -> None:
        """

        :param save_name:
        :param fig_size:
        :return:
        """
        fig, ax = plt.subplots(1, 1, figsize=fig_size)
        x = self.rawCounts.apply(np.log10).sort_values(ascending=False).reset_index()[0]
        x.plot(kind="line", ax=ax, lw=4)
        ax.set_xlabel("Barcodes sorted by frequency", fontsize=12)
        ax.set_ylabel("Log10 frequency", fontsize=12)
        plt.tight_layout()
        if save_name is not None:
            plt.savefig(save_name, dpi=300)
            plt.close()
        else:
            plt.show()



import editdistance
from tqdm import tqdm
from typing import Optional, Dict
import nmslib
import pandas as pd


class SeqCorrect:
    def __init__(
        self,
        min_counts: int = 20,
        max_dist: int = 3,
        ann_k: Optional[int] = None,
        ann_m: int = 30,
        ann_ef_construction: int = 100,
        ann_post: int = 0,
        n_threads: int = 2,
        ann_ef_search: int = 100,
        disable_progress_bar: bool = False,
    ):
        """
        Class for identifying erroneous barcodes and merging their frequency with their correct version.

        :param min_counts: Minimum barcode frequency threshold. Barcodes with less than this value are removed. This
                           check is performed on corrected counts.
        :param max_dist: Minimum distance threshold when comparing barcodes. Barcodes with less than this value is
                         considered to be an erroneous version of a higher abundance barcode.
        :param ann_k: Number of neighbours in the KNN search.
        :param ann_m: ANN index construction parameter M
        :param ann_ef_construction: ANN index construction EF
        :param ann_post: ANN index construction post
        :param n_threads: ANN number of threads
        :param ann_ef_search: ANN query parameter EF
        :param disable_progress_bar: If True, then progress bars are not shown
        """

        if ann_k is None:
            self.annK = 2 * ann_m  # Same as the default behaviour of nmslib
        else:
            self.annK = ann_k
        self.minCounts = min_counts
        self.maxDist = max_dist
        self.indexParams = {
            "M": ann_m,
            "efConstruction": ann_ef_construction,
            "post": ann_post,
            "indexThreadQty": n_threads,
        }
        self.queryParams = {"efSearch": ann_ef_search}
        self.index = None
        self.correctedCounts: Optional[pd.Series] = None
        self.disablePb = disable_progress_bar

    def _build_index(self, raw_counts: Dict[str, int]) -> None:
        index = nmslib.init(
            method="hnsw",
            space="leven",
            data_type=nmslib.DataType.OBJECT_AS_STRING,
            dtype=nmslib.DistType.INT,
        )
        index.addDataPointBatch(list(raw_counts.keys()))
        index.createIndex(self.indexParams, print_progress=True)
        self.index = index

    def _query_index(self, raw_counts: Dict[str, int]) -> Dict[str, str]:
        bc_list = list(raw_counts.keys())
        self.index.setQueryTimeParams(self.queryParams)
        nbrs = self.index.knnQueryBatch(
            bc_list, k=self.annK, num_threads=self.indexParams["indexThreadQty"]
        )
        nm = {}
        for n, i in tqdm(
            enumerate(nbrs),
            total=len(nbrs),
            desc="Querying ANN index",
            disable=self.disablePb,
        ):
            a = bc_list[n]
            for j in i[0]:
                b = bc_list[j]
                if j == n:
                    continue
                if raw_counts[a] > raw_counts[b]:
                    continue
                d = editdistance.eval(a, b)
                if d > self.maxDist:
                    break
                if a in nm:
                    if raw_counts[b] > raw_counts[nm[a]]:
                        nm[a] = b
                else:
                    nm[a] = b
        return nm

    def _correct(
        self, raw_counts: Dict[str, int], neighbors: Dict[str, str]
    ) -> Dict[str, int]:
        bc_list = list(raw_counts.keys())
        cor_counts = {}
        for i in tqdm(bc_list, desc="Merging barcodes", disable=self.disablePb):
            if i in neighbors:
                a = neighbors[i]
            else:
                a = i
            if a not in cor_counts:
                cor_counts[a] = 0
            cor_counts[a] += raw_counts[i]

        clean_counts = {}
        for i in tqdm(
            cor_counts, desc="Filtering corrected barcodes", disable=self.disablePb
        ):
            if raw_counts[i] <= cor_counts[i] and cor_counts[i] > self.minCounts:
                clean_counts[i] = cor_counts[i]
        return clean_counts

    def run(self, raw_counts: pd.Series) -> None:
        """
        Get corrected and filtered barcodes.

        :param raw_counts: Row barcodes counts, in form of a dictionary.
        :return: Corrected barcode counts in form of a dictionary
        """
        counts = dict(raw_counts.to_dict())
        self._build_index(counts)
        neighbors = self._query_index(counts)
        self.correctedCounts = pd.Series(self._correct(counts, neighbors))

    def save_to_csv(self, file_name: str):
        """

        :param file_name:
        :return:
        """
        self.correctedCounts.to_csv(file_name, header=False)



from typing import Generator, Tuple
from .config import logger
import os
import glob

__all__ = ["glob_files"]


def glob_files(
    directory: str,
    read1_pattern: str = "R1",
    read2_pattern: str = "R2",
    file_extension: str = "fastq.gz",
) -> Generator[Tuple[str, str], None, None]:
    """

    :param directory:
    :param read1_pattern:
    :param read2_pattern:
    :param file_extension:
    :return:
    """
    if os.path.isdir(directory):
        logger.debug("Directory found!")
    else:
        logger.error(f"Directory '{directory}' not found!")
        return None
    samples1 = sorted(glob.glob(f"{directory}/*{read1_pattern}.{file_extension}"))
    samples2 = sorted(glob.glob(f"{directory}/*{read2_pattern}.{file_extension}"))
    try:
        assert len(samples1) == len(samples2)
    except AssertionError:
        logger.error(
            "Different number of read1 and read2 files found. "
            "Please check that `read1_pattern` and `read2_pattern` parameters are correct. "
            "Also check that you have an even number of correctly labelled files in the directory."
        )
        return None
    else:
        if len(samples1) == 0:
            logger.error(
                "No samples found. Check that directory and file extension are correct. "
                "Also that `read1_pattern` and `read2_pattern` parameter are correct. "
                "For example, for miSeq runs `read1_pattern` and `read2_pattern` could "
                "look like: 'R1_001' and 'R2_001' respectively."
            )
            return None
        print(f"INFO: {len(samples1)} samples found")
        for r1, r2 in zip(samples1, samples2):
            yield r1, r2



from .paired_extractor import BarcodeExtractor
from .sequence_corrector import SeqCorrect
from .batch_analyzer import BarcodeAnalyzer
from .utils import *
from importlib_metadata import version

try:
    __version__ = version("bartide")
except ImportError:
    print("Bartide is not installed", flush=True)



from setuptools import setup, find_packages

setup(
    name="bartide",
    version=open("VERSION").readline().rstrip("\n"),
    description="A Python package to extract, correct and analyze nucleotide barcodes from sequenced reads.",
    packages=find_packages(),
    author="Parashar Dhapola",
    author_email="parashar.dhapola@gmail.com",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/parashardhapola/bartide",
    include_package_data=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "matplotlib",
        "loguru",
        "editdistance",
        "tqdm",
        "pandas",
        "joblib",
        "seaborn",
        "upsetplot",
    ],
    extras_require={"dev": ["pytest", "pytest-pep8", "pytest-cov", "black"]},
    keywords=["Text Mining", "Barcode", "nucleotide", "sequencing", "deduplicate"],
)
